use bevy::prelude::*;
use bevy::math::Mat4;

/// creates a conventional projection matrix from frustum planes
/// 
/// returns a potentially off-axis projection matrix
pub fn make_projection_rh_from_frustum(left: f32, right: f32, bottom: f32, top: f32, near:f32, far:f32) -> Mat4
{
    // based on OpenSceneGraph / glFrustum implementation
    let a = (right + left) / (right - left);
    let b: f32 = (top + bottom) / (top - bottom);
    let c= if far.abs() > f32::MAX  { -1.0 } else { -(far + near) / (far - near)};
    let d = if far.abs() > f32::MAX { -2.0 * near } else { -2.0 * far * near / (far - near) };

    Mat4::from_cols( 
        Vec4::new(2.0 * near / (right-left),0.0,0.0,0.0), 
        Vec4::new(0.0, 2.0 *near / (top-bottom), 0.0,0.0),
        Vec4::new(a, b, c, -1.0),
        Vec4::new(0.0, 0.0, d, 0.0))
}


/// creates a projection from a frustum planes with a reversed depth mapped to [0..1]
pub fn make_projection_rh_from_frustum_reversed(left: f32, right: f32, bottom: f32, top: f32, z_near:f32, z_far:f32) -> Mat4
{
    assert!(z_near > 0.0 && z_far > 0.0);

    //
    // reversed z 0..1 projection based on https://thxforthefish.com/posts/reverse_z/
    //
    let a = (right + left) / (right - left);
    let b: f32 = (top + bottom) / (top - bottom);
    
    let c =  z_near / (z_far - z_near);
    let d =  z_far * z_near / (z_far - z_near);


    Mat4::from_cols( 
        Vec4::new(2.0 * z_near / (right-left),0.0,0.0,0.0), 
        Vec4::new(0.0, 2.0 * z_near / (top-bottom), 0.0,0.0),
        Vec4::new(a, b, c, -1.0),
        Vec4::new(0.0, 0.0, d, 0.0))
}


pub fn make_projection_rh_custom(fov_y: f32, aspect_ratio: f32, z_near: f32, z_far: f32) -> Mat4
{
    let tan_fovy = (fov_y * 0.5).tan(); // use half angle beta
    let right = tan_fovy * aspect_ratio * z_near;
    let left = -right;
    let top = tan_fovy * z_near;
    let bottom = -top;

    //make_projection_rh_from_frustum(left, right, bottom, top, z_near, z_far)

    make_projection_rh_from_frustum_reversed(left, right, bottom, top, z_near, z_far)
}


#[cfg(test)]
mod tests {
    use bevy::prelude::Mat4;

    use super::make_projection_rh_from_frustum;


    #[test]
    fn compare_projections() {

        // build an on-axis frustum
        let fovy = 33.0_f32;
        let aspect_ratio = 1.6666_f32;
        let z_near = 1.0_f32;
        let z_far = 1000.0_f32;

        let tan_fovy = (fovy * 0.5).to_radians().tan(); // use half angle beta
        let right = tan_fovy * aspect_ratio * z_near;
        let left = -right;
        let top = tan_fovy * z_near;
        let bottom = -top;

        let mat_frust = make_projection_rh_from_frustum(left, right, bottom, top, z_near, z_far);

        println!("mat 1 {:?}",mat_frust);

        let mat_pers = Mat4::perspective_rh_gl(fovy.to_radians(),aspect_ratio,z_near,z_far);

        println!("mat 2 {:?}",mat_pers);

        assert!(mat_frust.abs_diff_eq(mat_pers, f32::EPSILON));

    }

}    
