
use bevy::{prelude::*, math::bool, prelude::shape::Quad };
use rand::prelude::*;
use bevy::render::primitives::Plane;

enum EyePos {
    Left,
    Right
}


#[derive(Default,Component)]
pub struct ScreenInfo {
    pub name : String,          // main (to identify the screen)
    width : f32,            // 3.08    (full width in m)
    height : f32,           // 2.33    (full height in m)
    
    center : Vec3,          // 0.0 -1.15 -3.08 (mid point of screen in global coordinated - tracking!)
    normal : Vec3,          // 0.0 0.0 1.0      (orientation of front side)
    up: Vec3,               // 0.0 1.0 0.0      (vertical axis)
    horizontal: Vec3,       // right vector computed as orthonormal   
   
    top : Vec3,
    bottom : Vec3,
    left : Vec3,
    right : Vec3
}

impl ScreenInfo {

    pub fn new(name: &str) -> Self {
        Self {
            name: name.to_string(),
            width: 3.08,
            height: 2.33,
            center: Vec3 { x: 0.0, y: -1.15, z: -3.08 },
            normal: Vec3::Z,
            up: Vec3::Y,
            horizontal: Vec3::ZERO,            
            top: Vec3::ZERO, 
            bottom: Vec3::ZERO, 
            left: Vec3::ZERO,
            right: Vec3::ZERO,
        }
    }

    pub fn build_debug_geometry(&self,
        mut commands: Commands,
        mut meshes: ResMut<Assets<Mesh>>,
        mut materials: ResMut<Assets<StandardMaterial>>,
        )
    {
            // plane
        commands.spawn(PbrBundle {
        mesh: meshes.add(Mesh::from( Quad { size: Vec2 { x: self.width, y: self.height }, flip:false })),
        material: materials.add(Color::rgb(0.3, 0.5, 0.3).into()),
        ..default()
        });
    }

    pub fn update(&mut self) -> &mut Self {
        
        // compute right as up ^ forward
        self.horizontal = self.up.cross(self.normal);

        // points of screen
        self.top = self.center + self.up * (self.height * 0.5);
        self.bottom = self.center - self.up * (self.height * 0.5);
        self.left = self.center - self.horizontal * (self.height * 0.5);
        self.right = self.center + self.horizontal * (self.height * 0.5);

        self
        
    }

    fn intersection(
        &self,
        p1: Vec3,
        p2: Vec3) -> (bool,Vec3)
    {

        let u = self.normal.dot(self.center - p1) / self.normal.dot(p2 - p1);

        let intersection_point = p1 + ((p2 - p1) * u);

        // returns a tuple if the intersection point is in front and the actual point
        (u >= 0.0 && u <= 1.0,intersection_point)
    }


    fn normal_intersection(&self,
        p1: Vec3) -> (bool,Vec3)
    {
        let p2 = p1 - self.normal * 10.0;
        
        self.intersection(p1, p2)
    }
    
}
